##############################################################################
# 
# Zope Public License (ZPL) Version 1.0
# -------------------------------------
# 
# Copyright (c) Digital Creations.  All rights reserved.
# 
# This license has been certified as Open Source(tm).
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
# 
# 1. Redistributions in source code must retain the above copyright
#    notice, this list of conditions, and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright
#    notice, this list of conditions, and the following disclaimer in
#    the documentation and/or other materials provided with the
#    distribution.
# 
# 3. Digital Creations requests that attribution be given to Zope
#    in any manner possible. Zope includes a "Powered by Zope"
#    button that is installed by default. While it is not a license
#    violation to remove this button, it is requested that the
#    attribution remain. A significant investment has been put
#    into Zope, and this effort will continue if the Zope community
#    continues to grow. This is one way to assure that growth.
# 
# 4. All advertising materials and documentation mentioning
#    features derived from or use of this software must display
#    the following acknowledgement:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    In the event that the product being advertised includes an
#    intact Zope distribution (with copyright and license included)
#    then this clause is waived.
# 
# 5. Names associated with Zope or Digital Creations must not be used to
#    endorse or promote products derived from this software without
#    prior written permission from Digital Creations.
# 
# 6. Modified redistributions of any form whatsoever must retain
#    the following acknowledgment:
# 
#      "This product includes software developed by Digital Creations
#      for use in the Z Object Publishing Environment
#      (http://www.zope.org/)."
# 
#    Intact (re-)distributions of any official Zope release do not
#    require an external acknowledgement.
# 
# 7. Modifications are encouraged but must be packaged separately as
#    patches to official Zope releases.  Distributions that do not
#    clearly separate the patches from the original work must be clearly
#    labeled as unofficial distributions.  Modifications which do not
#    carry the name Zope may be packaged in any form, as long as they
#    conform to all of the clauses above.
# 
# 
# Disclaimer
# 
#   THIS SOFTWARE IS PROVIDED BY DIGITAL CREATIONS ``AS IS'' AND ANY
#   EXPRESSED OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
#   IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
#   PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL DIGITAL CREATIONS OR ITS
#   CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
#   SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
#   LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
#   USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
#   ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
#   OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
#   OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
#   SUCH DAMAGE.
# 
# 
# This software consists of contributions made by Digital Creations and
# many individuals on behalf of Digital Creations.  Specific
# attributions are listed in the accompanying credits file.
# 
##############################################################################
"""
Interbase ZODB Storage

$Id: InterbaseStorage.py,v 1.1.1.1 2000/06/11 06:05:53 chrism Exp $
"""

__version__='$Revision: 1.1.1.1 $'[11:-2]

from ZODB import POSException, BaseStorage
import kinterbasdb
import string, os, time, base64, sys, binascii, time
from ZODB.TimeStamp import TimeStamp
from cPickle import loads

dbiraw = kinterbasdb.dbi.dbiRaw

class RelationalStorageError(POSException.StorageError): pass

class RelationalStorageInitError(RelationalStorageError): pass

# For sanity's sake, we name the items in a data record

OID,SERIAL,PRE,STATUS,DATA,DATALEN,VERSION,NV,DATASERIAL = range(9)
DBIRAW = type(dbiraw(''))

# dirty-read, immediate-conflict-raise Interbase transaction parameters

dirtyread_tpb = kinterbasdb.isc_tpb_nowait +\
                kinterbasdb.isc_tpb_no_rec_version

class InterbaseStorage(BaseStorage.BaseStorage):
    """ Relational Storage class stores ZODB records and transaction
    data in a relational DB -- this is the Interbase variety """

    def __init__(self, conn, name="InterbaseStorage",
                 zodb_data="zodb_data", zodb_trans="zodb_trans",
                 zodb_pack="zodb_pack", debug=0):
        self._debug = debug
        if self._debug:
            print "__init__"
        self._c = conn
        self._tables = {'data': zodb_data,
                        'trans': zodb_trans,
                        'pack': zodb_pack}
        self._testConn()
        BaseStorage.BaseStorage.__init__(self, name)
        self._oid = self.GetLastOid()
        
    def __len__(self):
        """ Return number of current records in the ODB """
        if self._debug:
            print "__len__"
        return self.GetNumberOfCurrentRecords()
    
    def GetLastOid(self):
        """ Obtain the last ('highest') oid used in the DB.  If there
        are no records in the DB yet, return the 'root' OID. """
        if self._debug:
            print "GetLastOid"
        dcs = self._doCommittedSQL
        tables = self._tables
        cmd = """
            SELECT z_oid, z_serial
            FROM %(data)s """ % tables
        recs = dcs(cmd = cmd, fetch="all")
        oids = []
        for rec in recs:
            oid = a2b(rec[0])
            oids.append(oid)
        oids.sort()
        try:
            return oids[-1]
        except:
            return '\0\0\0\0\0\0\0\0'
            
    def close(self):
        """ Close the DB connection """
        if self._debug:
            print "_close"
        self._closeConn()

    def supportsVersions(self):
        if self._debug:
            print "supportsVersions"
        return 1

    def supportsUndo(self):
        if self._debug:
            print "supportsUndo"
        return 1

    def load(self, oid, version):
        """ If there is version data matching the version passed
        in to this method for the object we've asked for, return
        the most current object in that version for the oid requested.
        Else, return the most current nonversion object for the oid
        requested. If we fail entirely, report a RelationalStorageError. """
        debug = self._debug
        GetCurrent = self.GetCurrent
        GetHistorical = self.GetHistorical
        if debug:
            print "\nload"
        self._lock_acquire()
        try:
            current = GetCurrent(oid=oid, committedsql=1)
            if not current:
                error = "No current record for oid %s (%s)" %(oid,b2a(oid))
                raise RelationalStorageError, error
            data = current[DATA]
            serial = current[SERIAL]
            if current[VERSION]:
                if current[VERSION] != version:
                    if current[NV]:
                        if debug:
                            print "going for nonversion record for %s" % b2a(oid)
                        nonversion = GetHistorical(oid=oid,
                                                    serial=current[NV],
                                                    committedsql=1)
                        data = nonversion[DATA]
                        serial = nonversion[SERIAL]
            return data, serial
        finally:
            self._lock_release()
    
    def GetCurrent(self, oid='', committedsql=0):
        if self._debug:
            print "GetCurrent"
        try:
            dcs = self._doCommittedSQL
            ducs = self._doUncommittedSQL
            debug = self._debug
            tables = self._tables
            GetHistorical = self.GetHistorical
            if debug:
                print ("committedsql parameter to getcurrent is %s"
                       % committedsql)
                print "getcurrent wants oid %s" % b2a(oid)
            cmd = """
              SELECT z_oid, z_serial, z_pre, z_status, z_data, z_datalen,
                     z_version, z_nv, z_dataserial
              FROM %(data)s
              WHERE z_oid = ?
              AND z_status = 'c' """ % tables
            s_oid = b2a(oid)
            if committedsql:
                ob = dcs(cmd=cmd,params=(s_oid,),fetch="one",dirtyread=1)
            else:
                ob = ducs(cmd=cmd,params=(s_oid,),fetch="one", dirtyread=1)
            if not ob:
                return
            ob = list(ob)
            if not ob[DATA] and ob[DATASERIAL]:
                if debug:
                    print ("getcurrent passing %s"
                            "as serial to getcurrent for oid %s"
                           % (a2b(ob[DATASERIAL]),s_oid))
                actual = GetHistorical(oid=oid,
                                       serial=a2b(ob[DATASERIAL]),
                                       committedsql=committedsql)
                if not actual:
                    return
                ob[DATA] = actual[DATA]
            ob[OID] = a2b(ob[OID])
            ob[SERIAL] = a2b(ob[SERIAL])
            ob[PRE] = a2b(ob[PRE])
            try:
                ob[DATA] = ob[DATA].value
            except AttributeError:
                pass
            ob[NV] = a2b(ob[NV])
            ob[DATASERIAL] = a2b(ob[DATASERIAL])
            return tuple(ob)
        except:
            if self._debug:
                import traceback
                traceback.print_exc()
            raise
        
    def GetHistorical(self, oid='', serial='', committedsql=0):
        debug = self._debug
        dcs = self._doCommittedSQL
        ducs = self._doUncommittedSQL
        debug = self._debug
        tables = self._tables
        GetHistorical = self.GetHistorical
        if debug:
            print "GetHistorical"
            print "gethist wants oid %s,serial %s" % (b2a(oid),b2a(serial))
            print ("committedsql parameter to gethistorical is %s"
                   % committedsql)
        try:
            cmd = """
                SELECT z_oid, z_serial, z_pre, z_status, z_data, z_datalen,
                       z_version, z_nv, z_dataserial
                FROM zodb_data
                WHERE z_oid = ?
                AND z_serial = ? """
            s_oid, s_serial = map(b2a, (oid, serial))
            if committedsql:
                ob = dcs(cmd=cmd,params=(s_oid,s_serial),
                         fetch="one",dirtyread=1)
            else:
                ob = ducs(cmd=cmd,params=(s_oid,s_serial),
                          fetch="one",dirtyread=1)            
            if not ob:
                if debug:
                    print 'gethistorical could not retrieve %s' 
                return
            ob = list(ob)
            if debug:
                print "non data serial data ob has type %s" % type(ob[DATA])
            if (not ob[DATA]) and ob[DATASERIAL]:
                if debug:
                    print "going to dataserial record %s" % ob[DATASERIAL]
                if committedsql:
                    actual = dcs(cmd=cmd,params=(s_oid,ob[DATASERIAL]),
                                                 fetch="one",dirtyread=1)
                else:
                    actual = ducs(cmd=cmd,params=(s_oid,ob[DATASERIAL]),
                                                 fetch="one",dirtyread=1)
                if not actual:
                    return
                if debug:
                    print "dataserial ob data ob has type %s" % type(ob[DATA])
                ob[DATA] = actual[DATA]
            ob[OID] = a2b(ob[OID])
            ob[SERIAL] = a2b(ob[SERIAL])
            ob[PRE] = a2b(ob[PRE])
            ob[DATA] = ob[DATA].value
            ob[NV] = a2b(ob[NV])
            ob[DATASERIAL] = a2b(ob[DATASERIAL])
            return tuple(ob)
        except:
            if debug:
                import traceback
                traceback.print_exc()
            raise
        
    def store(self, oid, serial, p, version, transaction):
        """
        If the oid passed in to this method doesn't already
        have an 'old' current record in the DB, just go ahead
        and write it as a new object.  If there is an 'old' current
        record for this object, see if the 'old' current record has
        version data.  If the 'old' current record does have version
        data and the version that is passed in to this method doesn't
        match the 'old' current record's version, raise a
        VersionLockError (an object cannot be locked in two versions
        at the same time).  If the 'old' current record has version data,
        and the version data contains references to a nonversion record,
        grab the info about the nonversion record and use it in the
        insert as the nonversion record info for the new object.  If
        the 'old' current record does *not* have version data, use
        info about the 'old' current record as nonversion data.  Then
        write the new object out to a record.  Return the serial number
        (transaction id) used to write the object.  The SQL calls in this
        method are not committed until the '_finish' method is called,
        thus if a failure occurs, the database will not be left in an
        inconsistent state.
        """
        if transaction is not self._transaction:
            raise POSException.StorageTransactionError(self, transaction)

        self._lock_acquire()
        try:
            try:
                debug = self._debug
                if debug:
                    print "\nstore"            
                InsertObject = self.InsertObject
                SetStatus = self.SetStatus
                try:
                    current = self.GetCurrent(oid=oid)
                except kinterbasdb.kinterbasError, error:
                    if debug:
                        import traceback
                        traceback.print_exc()
                    raise POSException.ConflictError, (serial, oid, error)
                    # ok.  first of all, this isn't even necessary right now
                    # for ZODB3.  In future versions of the ZODB, it might be.
                    # When calls to the ZODB become nonserialized (ZODB4), this
                    # exception will prevent database corruption.
                    # when establishing the Interbase database
                    # connection, we have set the isc_tpb_nowait
                    # and the isc_tpb_no_rec_version TPB parameters within
                    # the connection init.
                    # We do this in order to catch reads to 'dirty' rows.
                    # To properly lock a row representing an object,
                    # we want to catch reads to rows that another
                    # connection has updated (e.g. in the case that a
                    # connection in another transaction has updated but
                    # not committed the 'current' record while changing its
                    # 'z_status'  we'll get back an exception).
                    # We raise a conflict error so the upper layer can
                    # retry.  This happens ad inifinitum until the
                    # transaction that caused the dirty read commits or
                    # until the upper layer backs off and bails.
                    # Note that we only need to do this because
                    # we're depending on the DB telling us what the current
                    # record is.  If we used in-memory indexes that are
                    # thread-shared for oid currency, we wouldn't need to do
                    # this as Jim's BaseStorage thread locking facility would
                    # do all the hard work.
                    #
                    # The advantage to this scheme is that our "current"
                    # set of records is always consistent and can never
                    # become corrupted by a transaction failure.
                    # The disadvantage to this scheme is that it
                    # is probably much slower in many cases.

                newserial = self._serial
                if not current:
                    InsertObject(oid=oid, serial=newserial, pre="",
                                 status="c", data=p, datalen=len(p),
                                 version=version, nv="", dataserial="")
                    return newserial
                old_version = current[VERSION]
                if old_version and old_version != version:
                    raise POSException.VersionLockError, oid
                if current[SERIAL] != serial:
                    raise POSException.ConflictError, (serial, oid)

                pre = current[SERIAL]
                SetStatus(oid=oid, serial=serial, status="h")
                if version:
                    if old_version:
                        nv=current[NV]
                    else:
                        nv = pre
                else:
                    nv = ''
                InsertObject(oid=oid, serial=newserial, pre=pre, status="c",
                             data=p, datalen=len(p), version=version, nv=nv,
                             dataserial="")
                return newserial
            except:
                import traceback
                traceback.print_exc()
                raise
        finally:
            self._lock_release()

    def InsertObject(self, oid, serial, pre, status, data, datalen,
                     version, nv, dataserial):
        if self._debug:
            print "InsertObject"
        tables = self._tables
        ducs = self._doUncommittedSQL
        cmd = """
        INSERT INTO %(data)s (z_oid, z_serial, z_pre, z_status, z_data,
                              z_datalen, z_version, z_nv, z_dataserial)
        VALUES (?, ?, ?, ?, ?, ?, ?, ?, ?)""" % tables
        data = dbiraw(data) # interbaseism
        s_oid = b2a(oid)
        s_serial = b2a(serial)
        s_pre = b2a(pre)
        s_nv = b2a(nv)
        s_dataserial = b2a(dataserial)
        row = (s_oid, s_serial, s_pre, status, data, datalen,
               version, s_nv, s_dataserial)
        ducs(cmd = cmd, params=row)

    def SetStatus(self, oid, serial, status):
        if self._debug:
            print "SetStatus"
        tables = self._tables
        ducs = self._doUncommittedSQL
        cmd = """
        UPDATE %(data)s
        SET z_status = ?
        WHERE z_oid = ?
        AND z_serial = ?""" % tables
        row = (status, b2a(oid), b2a(serial))
        ducs(cmd = cmd, params = row)
        
    def undo(self, transaction_id):
        """
        Regular nonundoable undo.

        Begin a SQL transaction.  If there's an error at any point,
        abort the transaction.

        Get all the records out of the DB that correspond to the
        passed-in transaction id (serial #).  If there are no
        records corresponding to the passed-in transaction id,
        raise an UndoError.  If there are records, loop through
        them one-by-one.  If the record corresponding to the passed-in
        transaction id in the DB doesn't match the 'current'
        record for an object (if its serial number differs),
        the object has been rewritten in a later transaction and cannot
        be undone, in which case we raise an UndoError which
        aborts this transaction.

        Loop through the records corresponding to the passed-in
        transaction id a second time.  If a record has 'previous'
        data (a pointer to an older version of the object), get the
        'previous' record and make it the 'current' record.  Keep
        tabs on the records we do this to and append their oids to
        a list for later return.  If the record has 'previous' data,
        and it also has associated version data, replace the 'current'
        record's version data with the previous record's version data.
        If the record has no 'previous' data, delete it from the 'current'
        index and delete its version data from the DB.  Delete all records
        from the DB having to do with the passed-in transaction id.

        Commit the SQL transaction.  Return the list of OIDs that had
        'previous' data so the layer above can reload thosse objects.
        """
        self._lock_acquire()
        if self._debug:
            print "\undo"
        try:
            try:
                UndoError = POSException.UndoError
                self._beginSQL()
                if not self.TransactionExists(transaction_id):
                    err = "Invalid Undo Transaction Id %s" % transaction_id 
                    raise UndoError, err
                if not self.TransactionIsUndoable(transaction_id):
                    err = "Non-undoable transaction %s" % transaction_id
                    raise UndoError, err
                oids = self.GetCurrentOidsWithAPre(transaction_id)
                self.MarkOldRecordsAsCurrent(transaction_id)
                self.DeleteDataRecords(transaction_id)
                self.DeleteTransactionRecord(transaction_id)
            except:
                self._abortSQL()
                type, value = sys.exc_info()[:2]
                import traceback
                traceback.print_exc()
            else:
                self._commitSQL()
                if self._debug:
                    print "oids returned from undo were %s" % (map(b2a, oids))
                return oids
        finally:
            self._lock_release()

    def TransactionExists(self, t_id):
        if self._debug:
            print "TransactionExists"
        tables = self._tables
        cmd = """
        SELECT z_serial
        FROM %(trans)s
        WHERE z_serial = ?""" % tables
        r=not not self._doUncommittedSQL(cmd=cmd,params=(t_id,),fetch="all",
                                         debug=self._debug)
        return r

    def TransactionIsUndoable(self, t_id):
        if self._debug:
            print "TransactionIsUndoable"
        tables = self._tables
        cmd = """
              SELECT COUNT(*)
              FROM %(data)s
              WHERE z_serial = ?
              AND z_status != 'c' """ % tables
        r = self._doUncommittedSQL(cmd=cmd,params=(t_id,),fetch="all",
                                   debug=self._debug)
        count = r[0]
        return count

    def MarkOldRecordsAsCurrent(self, t_id):
        """ Used by regular undo.  This should be an UPDATE WHERE EXISTS. """
        if self._debug:
            print "MarkOldRecordsAsCurrent"
        tables = self._tables
        cmd = """
            SELECT * FROM %(data)s
            WHERE z_serial = ?
            AND z_status IN ('c', 'g') """ % tables
        recs = self._doUncommittedSQL(cmd=cmd,params=(t_id,),
                                      fetch="all", debug=self._debug)
        for rec in recs:
            if rec[PRE]:
                cmd = """
                    UPDATE %(data)s SET z_status = 'c'
                    WHERE z_serial = ?
                    AND z_oid = ?""" % tables
                self._doUncommittedSQL(cmd=cmd,params=(rec[PRE],rec[OID]),
                                       debug=self._debug)

        # vestigial, but this is the way it *should* go
        #
        #cmd = """
        #    UPDATE %(data)s targ
        #    SET z_status = 'c'
        #    WHERE EXISTS ( SELECT * FROM %(data)s curr WHERE
        #                   curr.z_oid = targ.z_oid AND
        #                   curr.z_pre = targ.z_serial AND
        #                   curr.z_status = 'c' AND
        #                   curr.z_serial = ? ) """ % tables
        #self._doUncommittedSQL(cmd=cmd,params=(t_id,), debug=self._debug)
        
    def DeleteDataRecords(self, t_id):
        if self._debug:
            print "MarkCurrentRecordsAsHistorical"
        tables = self._tables
        cmd = """
            DELETE FROM %(data)s WHERE z_serial = ?
            """ % tables
        recs = self._doUncommittedSQL(cmd=cmd,params=(t_id,),debug=self._debug)

    def DeleteTransactionRecord(self, t_id):
        if self._debug:
            print "DeleteTransactionRecord"
        tables = self._tables
        cmd = """
             DELETE FROM %(trans)s
             WHERE z_serial = ? """ % tables
        self._doUncommittedSQL(cmd=cmd,params=(t_id,),debug=self._debug)
        
    def GetCurrentOidsWithAPre(self, t_id):
        if self._debug:
            print "GetCurrentOidsWithAPre"
        tables = self._tables
        cmd = """
              SELECT z_oid
              FROM %(data)s
              WHERE z_serial = ?
              AND z_status IN ('c', 'g') 
              AND z_pre != '' """ % tables

        recs = self._doUncommittedSQL(cmd=cmd, params=(t_id,),
                                      fetch = "all", debug=self._debug)
        oids = []
        for oid in recs:
            oid = oid[0]
            if oid is not None:
                oids.append(a2b(oid))
        return oids
    
    def undoableundo(self, transaction_id):
        """

        THIS DOESNT YET WORK!!

        waiting for ZODB4 transaction support for undo.

        """
        self._lock_acquire()
        if self._debug:
            print "undo"
        try:
            try:
                UndoError = POSException.UndoError
                self._beginSQL()
                if self._debug:
                    print "doing transactionexists"
                if not self.TransactionExists(transaction_id):
                    err = "Invalid Undo Transaction Id %s" % transaction_id 
                    raise UndoError, err
                if self._debug:
                    print "doing transactionundoable"
                if not self.TransactionIsUndoable(transaction_id):
                    err = "Non-undoable transaction %s" % transaction_id
                    raise UndoError, err
                t=time.time()
                t=apply(TimeStamp,(time.gmtime(t)[:5]+(t%60,)))
                serial=`t`
                if self._debug:
                    print "undoing transaction %s" % transaction_id
                    print ("serial created for undo transaction is %s"
                           % b2a(serial))
                    print "doing makedummyrecords"
                self.MakeDummyUndoRecords(serial, transaction_id)
                if self._debug:
                    print "doing undopopulatedrecords"
                self.UndoPopulatedRecords(serial, transaction_id)
                if self._debug:
                    print "doing undounpopulatedrecords"
                self.UndoUnpopulatedRecords(serial, transaction_id)
                if self._debug:
                    print "doing clearundostatus"
                self.MarkCurrentRecordsAsOld(transaction_id)    
                if self._debug:
                    print "doing getcurrentids"
                self.InsertTrans(serial, 'undo', 'Zope', '')
                oids = self.GetCurrentOids(serial)
            except:
                self._abortSQL()
                type, value = sys.exc_info()[:2]
                import traceback
                traceback.print_exc()
            else:
                self._commitSQL()
                if self._debug:
                    print "oids returned were %s" % (map(b2a, oids))
                return oids
        finally:
            self._lock_release()

    def MarkCurrentRecordsAsOld(self, t_id):
        """ Used only by the yet unimplementable undoable undo. """
        if self._debug:
            print "ClearUndoStatus"
        tables = self._tables
        cmd = """
            UPDATE %(data)s
            SET z_status = 'h'
            WHERE z_serial = ? """ % tables
        self._doUncommittedSQL(cmd=cmd,params=(t_id,), debug=self._debug)

    def MakeDummyUndoRecords (self, serial, t_id):
        """ Used only by the yet unimplementable undoable undo. """
        if self._debug:
            print "MakeDummyUndoRecords"
        tables = self._tables
        cmd = """
              INSERT INTO %(data)s
              (z_oid, z_serial, z_pre, z_status, z_data, z_datalen,
              z_version, z_nv, z_dataserial)
              SELECT z_oid, ?, z_serial, 'g', NULL, 0, z_version, z_nv, ''
              FROM %(data)s
              WHERE z_serial = ?
              AND z_pre = '' """ % tables
        self._doUncommittedSQL(cmd=cmd,params=(b2a(serial), t_id),
                               debug=self._debug)

    def UndoUnpopulatedRecords(self, serial, t_id):
        """ This undoes an undo or undoes a version commit/abort
            Used only by the yet-unimplementable undoableundo """
        if self._debug:
            print "UndoPopulatedRecords"
        tables = self._tables
        cmd = """
              INSERT INTO %(data)s
               (z_oid, z_serial, z_pre, z_status, z_datalen, z_version,
               z_nv, z_dataserial)
              SELECT pre.z_oid, ?, rec.z_serial, 'c', pre.z_datalen,
              pre.z_version, pre.z_nv, pre.z_serial
              FROM %(data)s rec, %(data)s pre
              WHERE rec.z_serial = ?
              AND rec.z_dataserial != ''
              AND rec.z_data IS NULL
              AND rec.z_status != 'g' /* Chris added */
              AND rec.z_pre = pre.z_serial
              AND rec.z_oid = pre.z_oid """ % tables
        self._doUncommittedSQL(cmd=cmd,params=(b2a(serial), t_id),
                               debug=self._debug)

    def UndoPopulatedRecords(self, serial, t_id):
        """ Used only by the yet unimplementable undoable undo. """
        if self._debug:
            print "MakeUndoUnpopulatedRecords"
        tables = self._tables
        cmd = """
              INSERT INTO %(data)s
               (z_oid, z_serial, z_pre, z_status, z_datalen, z_version,
               z_nv, z_dataserial)
              SELECT pre.z_oid, ?, rec.z_serial, 'c', pre.z_datalen,
              pre.z_version, pre.z_nv, pre.z_serial
              FROM %(data)s rec, %(data)s pre
              WHERE rec.z_serial = ?
              AND pre.z_dataserial = ''
              AND rec.z_data IS NOT NULL
              AND rec.z_status != 'g'
              AND rec.z_pre = pre.z_serial
              AND rec.z_oid = pre.z_oid """ % tables
        self._doUncommittedSQL(cmd=cmd,params=(b2a(serial), t_id),
                               debug=self._debug)

    def GetCurrentOids(self, serial):
        if self._debug:
            print "GetCurrentOids"
        tables = self._tables
        cmd = """
              SELECT z_oid
              FROM %(data)s
              WHERE z_serial = ?
              AND z_status = 'c' """ % tables

        recs = self._doUncommittedSQL(cmd=cmd, params=(b2a(serial),),
                                      fetch = "all", debug=self._debug)
        oids = []
        for oid in recs:
            oid = oid[0]
            if oid is not None:
                oids.append(a2b(oid))
        return oids

    def GetNumberOfCurrentRecords(self):
        if self._debug:
            print "GetNumberOfCurrentRecords"
        tables = self._tables
        cmd = """
              SELECT COUNT(*)
              FROM %(data)s
              WHERE z_status = 'c' """ % tables
        num = self._doCommittedSQL(cmd=cmd, fetch = "one", debug=self._debug)
        return int(num[0])
    
    def undoLog(self, first, last, filter=None):
        """
        Retrieve all current transactions from the DB that have not been
        packed.  Populate a dictionary with return values (id, time,
        user_name, and description) culled from the transaction data.  The
        transactions returned may be limited to a subset of
        available transactions through the 'first' and 'last' integer
        values passed in (think 'slice'), and the resulting dictionary
        may be filtered by the 'filter' function passed in.  A
        (seldom used) 'extended' pickle stored in each record may also
        perform data manipulation on the returned dictionary.
        """
        self._lock_acquire()
        if self._debug:
            print "undoLog"
        try:
            dcs = self._doCommittedSQL
            tables = self._tables
            cmd = ("SELECT z_serial, z_status, z_user, z_desc, z_ext"
                   " FROM %(trans)s" % tables)
            recs = dcs(cmd = cmd, fetch="all")
            pos = len(recs)
            encode = base64.encodestring
            l = []
            append = l.append
            i = 0
            while i < last and pos:
                pos = pos - 1
                if i < first:
                    i = i + 1
                    continue
                s_serial, status, user, desc, s_ext = recs[pos]
                serial = a2b(s_serial)
                d = {'id': encode(serial)[:-1],
                     'time': TimeStamp(serial).timeTime(),
                     'user_name': user,
                     'description': desc}
                if s_ext:
                    if type(s_ext) == DBIRAW:
                        if len(s_ext.value) != 0:
                            d.update(loads(s_ext.value))

                if filter is None or filter(d):
                    append(d)
                    i = i + 1
            return l
        
        finally:
            self._lock_release()

    def modifiedInVersion(self, oid):
        """
        If the current record referenced by the oid passed in has
        version data, return the version name that the object is
        partipating in.  If it does not, return a blank string.
        """
        debug = self._debug
        GetCurrent = self.GetCurrent
        self._lock_acquire()
        if debug:
            print "modifiedInVersion"
        try:
            ob = GetCurrent(oid=oid, committedsql=1)
            if ob[VERSION]:
                if debug:
                    print ("%s was modified in version %s" %
                           (b2a(oid), ob[VERSION]))
                    return ob[VERSION]
            if debug:
                print ("No xcept, %s was not modified in a version"
                       % b2a(oid))
            return ''
        finally:
            self._lock_release()

    def abortVersion(self, src, transaction):
        """
        Abort the version referenced by the 'src' parameter passed in,
        which is a version name.  To do so, get all DB records with
        a version name matching 'src'.  If there are no records matching,
        return.  If there are records, iterate through the records
        one-by-one.  For each record, if the record has nonversion
        information, append its oid to a list to be returned later and
        store the record referenced by the nonversion info in the DB.  If
        the record does not have nonversion information, delete it.
        Return all oids of objects that contained  nonversion information.
        The transaction which aborts this version is undoable.
        """
        if self._debug:
            print "abortVersion"
        if transaction is not self._transaction:
            raise POSException.StorageTransactionError(self, transaction)
        self._lock_acquire()
        try:
            try:
                oids = self.GetCurrentOidsInVersion(src)
                if not oids:
                    return
                self.InsertAbortedVersionAsNew(self._serial, src)
                self.MakeCurrentVersionRecordsHistorical(src)
                return oids
            except:
                if self._debug:
                    import traceback
                    traceback.print_exc()
                raise
        finally:
            self._lock_release()
            
    def commitVersion(self, src, dest, transaction):
        """
        Commit the version records from the 'src' version to the 'dest'
        version.  If the 'dest' version is null or None, commit the
        objects modified in the version to the 'current' (nonversion)
        recordset.  In this implementation, dest is ignored and
        the version is always committed to the 'current' recordset.
        """
        if self._debug:
            print "commitVersion"
        if transaction is not self._transaction:
            raise POSException.StorageTransactionError(self, transaction)
        self._lock_acquire()
        try:
            try:
                if dest:
                    raise 'Unimplemented committing between versions'
                oids = self.GetCurrentOidsInVersion(src)
                if not oids:
                    return
                self.InsertCommittedVersionAsNew(self._serial, src)
                self.MakeCurrentVersionRecordsHistorical(src)
                return oids
            except:
                import traceback
                traceback.print_exc()
                raise
        finally:
            self._lock_release()
                
    def GetCurrentOidsInVersion(self, version):
        if self._debug:
            print "GetCurrentOidsInVersion"
        tables = self._tables
        oids = []
        oappend = oids.append
        cmd = """
            SELECT z_oid
            FROM %(data)s
            WHERE z_version = ?
            AND z_status = 'c' """ % tables
        roid = self._doUncommittedSQL(cmd=cmd,params=(version,),fetch="all")
        for tup in roid:
            oappend(a2b(tup[0]))
        if self._debug:
            print "getcurrentoidsinversion returned %s" % (oids)
        return oids

    def InsertCommittedVersionAsNew(self, serial, version):
        """ This should be two insert selects.  But the insert select
        doesn't work, and parameterized queries dont work in the select
        either for some reason. """
        if self._debug:
            print "InsertCommittedVersionAsNew"
        tables = self._tables
        s_serial = b2a(serial)
        cmd = """ SELECT z_oid, %s, z_serial, 'c', 0, '', '',
              z_dataserial
              FROM %s
              WHERE z_version = %s
              AND z_status = 'c'
              AND z_dataserial != '' """ % (_sql_quote(s_serial),
                                            tables['data'],
                                            _sql_quote(version))
        rows = self._doUncommittedSQL(cmd=cmd,fetch='all', debug=self._debug)

        for row in rows:
            cmd = """ INSERT INTO %(data)s
                 (z_oid, z_serial, z_pre, z_status, z_datalen, z_version, z_nv,
                 z_dataserial) values (?, ?, ?, ?, ?, ?, ?, ?) """ % tables
            self._doUncommittedSQL(cmd=cmd,params=row,debug=self._debug)

        cmd = """ SELECT z_oid, %s, z_serial, 'c', 0, '', '',
              z_serial
              FROM %s
              WHERE z_version = %s
              AND z_status = 'c'
              AND z_data IS NOT NULL
              AND z_dataserial = '' """ % (_sql_quote(s_serial),
                                           tables['data'],
                                           _sql_quote(version))
        rows = self._doUncommittedSQL(cmd=cmd, fetch="all", debug=self._debug)
        for row in rows:
            cmd = """ INSERT INTO %(data)s
                 (z_oid, z_serial, z_pre, z_status, z_datalen, z_version, z_nv,
                 z_dataserial) values (?, ?, ?, ?, ?, ?, ?, ?) """ % tables
            self._doUncommittedSQL(cmd=cmd,params=row,debug=self._debug)
            
    def InsertAbortedVersionAsNew(self, serial, version):
        """ This should use insert selects, I couldn't get it
        working properly, though. """
        if self._debug:
            print "InsertAbortedVersionAsNew"
        tables = self._tables
        s_serial = b2a(serial)
        cmd = """
            SELECT rec.z_oid, %s, rec.z_serial, 'c', 0, '', '', nv.z_dataserial
            FROM %s rec, %s nv
            WHERE rec.z_version = %s
            AND rec.z_status = 'c'
            AND rec.z_data IS NULL
            AND rec.z_dataserial != ''
            AND rec.z_nv = nv.z_serial
            AND rec.z_oid = nv.z_oid """ % (_sql_quote(s_serial),
                                            tables['data'],
                                            tables['data'],
                                            _sql_quote(version))
        rows = self._doUncommittedSQL(cmd=cmd, fetch='all', debug=self._debug)

        for row in rows:
            cmd = """ INSERT INTO %(data)s
                 (z_oid, z_serial, z_pre, z_status, z_datalen, z_version, z_nv,
                 z_dataserial) values (?, ?, ?, ?, ?, ?, ?, ?) """ % tables
            self._doUncommittedSQL(cmd=cmd,params=row,debug=self._debug)

        cmd = """
              SELECT rec.z_oid, %s, rec.z_serial,'c',0,'','', nv.z_serial
              FROM %s rec, %s nv
              WHERE rec.z_version = %s
              AND rec.z_status = 'c'
              AND rec.z_data IS NOT NULL
              AND rec.z_dataserial = ''
              AND rec.z_nv = nv.z_serial
              AND rec.z_oid = nv.z_oid """ % (_sql_quote(s_serial),
                                            tables['data'],
                                            tables['data'],
                                            _sql_quote(version))
        rows = self._doUncommittedSQL(cmd=cmd, fetch='all', debug=self._debug)
        
        for row in rows:
            cmd = """ INSERT INTO %(data)s
                 (z_oid, z_serial, z_pre, z_status, z_datalen, z_version, z_nv,
                 z_dataserial) values (?, ?, ?, ?, ?, ?, ?, ?) """ % tables
            self._doUncommittedSQL(cmd=cmd,params=row,debug=self._debug)

        cmd = """
              SELECT z_oid, %s, z_serial,'g',0,'','', ''
              FROM %s
              WHERE z_version = %s
              AND z_status = 'c'
              AND z_nv = '' """ % (_sql_quote(s_serial),
                                   tables['data'],
                                   _sql_quote(version))
        
        rows = self._doUncommittedSQL(cmd=cmd, fetch='all', debug=self._debug)
        
        for row in rows:
            cmd = """ INSERT INTO %(data)s
                 (z_oid, z_serial, z_pre, z_status, z_datalen, z_version, z_nv,
                 z_dataserial) values (?, ?, ?, ?, ?, ?, ?, ?) """ % tables
            self._doUncommittedSQL(cmd=cmd,params=row,debug=self._debug)

    def FancyNonworkingInsertCommittedVersionAsNew(self, serial, version):
        """ Tried to get fancy, wasn't working, other one worked. """
        cmd = """
            INSERT INTO %s
            (z_oid, z_serial, z_pre, z_status, z_datalen, z_version, z_nv,
             z_dataserial)
            SELECT z_oid, %s, z_serial, 'c', '', '', '', z_dataserial
            FROM %s
            WHERE z_version = %s
            AND z_status = 'c'
            AND z_dataserial != ''
            AND z_data IS NULL """ % (tables['data'],_sql_quote(s_serial),
                                          tables['data'], _sql_quote(version))
        self._doUncommittedSQL(cmd=cmd,debug=1)
        cmd = """
            INSERT INTO %s
            (z_oid, z_serial, z_pre, z_status, z_datalen, z_version, z_nv,
             z_dataserial)
            SELECT z_oid, %s, z_serial, 'c', z_datalen, '', '', z_serial
            FROM %s
            WHERE z_version = %s
            AND z_status = 'c'
            AND z_dataserial = ''""" % (tables['data'],_sql_quote(s_serial),
                                        tables['data'], _sql_quote(version))
        self._doUncommittedSQL(cmd=cmd,debug=1)

    def FancyNonworkingInsertAbortedVersionAsNew(self, serial, version):
        """ Tried to get fancy, wasnt working, other one worked """
        if self._debug:
            print "InsertAbortedVersionAsNew"
        tables = self._tables
        s_serial = b2a(serial)
        cmd = """
            INSERT INTO %(data)s
            (z_oid, z_serial, z_pre, z_status, z_datalen, z_version, z_nv,
             z_dataserial)
            SELECT rec.z_oid, ?, rec.z_serial,'c',nv.z_datalen,'','',
                   nv.z_dataserial
            FROM %(data)s rec, %(data)s nv
            WHERE rec.z_version = ?
            AND rec.z_status = 'c'
            AND rec.z_data IS NULL
            AND rec.z_dataserial != ''
            AND rec.z_nv = nv.z_serial
            AND rec.z_oid = nv.z_oid """ % tables
        self._doUncommittedSQL(cmd=cmd,params=(s_serial,version))
        cmd = """
            INSERT INTO %(data)s
            (z_oid, z_serial, z_pre, z_status, z_datalen, z_version, z_nv,
             z_dataserial)
            SELECT rec.z_oid,?,rec.z_serial,'c',nv.z_datalen,'','',
                   nv.z_serial
            FROM %(data)s rec, %(data)s nv
            WHERE rec.z_version = ?
            AND rec.z_status = 'c'
            AND rec.z_data IS NOT NULL
            AND rec.z_dataserial = ''
            AND rec.z_nv = nv.z_serial
            AND rec.z_oid = nv.z_oid """ % tables
        self._doUncommittedSQL(cmd=cmd,params=(s_serial,version))
        
    def MakeCurrentVersionRecordsHistorical(self, version):
        if self._debug:
            print "MakeCurrentVersionRecordsHistorical"
        try:
            tables = self._tables
            cmd = """
                UPDATE %(data)s
                SET z_status = 'h'
                WHERE z_version = ?
                AND z_status = 'c' """ % tables
            self._doUncommittedSQL(cmd=cmd,params=(version,))
        except:
            import traceback
            traceback.print_exc()
            raise
        
    def versionEmpty(self, version):
        if self._debug:
            print "versionEmpty"
        tables = self._tables
        cmd = """
            SELECT * FROM %(data)s
            WHERE z_version = ?
            AND z_status = 'c'""" % tables
        return not self._doCommittedSQL(cmd=cmd,params=(version,),fetch="one")

    def versions(self, max=None):
        if self._debug:
            print "versions"
        dcs = self._doCommittedSQL
        cmd = """
            SELECT DISTINCT z_version FROM %(data)s """ % tables
        recs = dcs(cmd=cmd,fetch="all")
        l = []
        a = l.append
        max = max or len(recs)
        for version in recs[:max]:
            if self.versionEmpty(version): continue
            a(version)
        return l

    def history(self, oid, version=None, size=1, filter=None):
        if self._debug:
            print "history"
        raise 'Unimplemented'
    
    def _begin(self, tid, u, d, e):
        debug = self._debug
        _beginSQL = self._beginSQL
        if debug:
            print "\n_begin"
            print ''
        _beginSQL()

    def _finish(self, tid, u, d, e):
        debug = self._debug
        _abortSQL = self._abortSQL
        _commitSQL = self._commitSQL
        InsertTrans = self.InsertTrans
        if debug:
            print "_finish\n", u, d
        try:
            serial = self._serial
            InsertTrans(serial, u, d, e)
        except:
            _abortSQL()
            type, value = sys.exc_info()[:2]
            raise type, value
        else:
            _commitSQL()

    def InsertTrans(self, serial, u, d, e):
        debug = self._debug
        ducs = self._doUncommittedSQL
        tables = self._tables
        if debug:
            print "InsertTrans"
        s_ext = dbiraw(e)
        s_serial = b2a(serial)
        cmd = """
        INSERT INTO %(trans)s
        (z_serial, z_user, z_desc, z_ext)
        VALUES (?, ?, ?, ?)""" % tables
        ducs(cmd = cmd, params = (s_serial,u,d,s_ext))
        
    def pack(self, t, referencesf):
        if self._debug:
            print "pack"
        ducs = self._doUncommittedSQL
        dcs = self._doCommittedSQL
        tables = self._tables
        if self._debug:
            print "acquiring lock"
        self._lock_acquire()
        try:
            try:
                self._beginSQL()
                if self._debug:
                    print "getting stop time"
                stop = `apply(TimeStamp, time.gmtime(t)[:5]+(t%60,))`
                if self._debug:
                    print "building pack table"
                cmd = """ INSERT INTO %(pack)s (z_oid, z_serial)
                       SELECT z_oid, MAX(z_serial)
                       FROM %(data)s
                       WHERE z_serial < ?
                       GROUP BY z_oid""" % tables
                ducs(cmd=cmd,params=(b2a(stop),), debug=self._debug)
                if self._debug:
                    print "removing unpackable records from pack table"
                rootl = ['\0\0\0\0\0\0\0\0']
                pop = rootl.pop
                while rootl != []:
                    oid = pop()
                    s_oid = b2a(oid)
                    cmd = """SELECT z_data
                             FROM %(data)s WHERE z_oid = ?
                             AND z_serial = ( SELECT z_serial
                                              FROM %(pack)s
                                              WHERE z_oid = ? )""" % tables
                    p = ducs(cmd=cmd,params=(s_oid,s_oid),fetch="one",
                             debug=self._debug)
                    if p is None:
                        continue
                    if p[0] is None:
                        continue
                    data = p[0].value #interbaseism
                    referencesf(data, rootl)
                    cmd = """ DELETE FROM %(pack)s
                              WHERE z_oid = ?""" % tables
                    ducs(cmd=cmd,params=(s_oid,),debug=self._debug)
                cmd = """ DELETE FROM %(data)s WHERE z_oid IN
                          (SELECT z_oid FROM zodb_pack)
                          AND z_serial < ? """ % tables
                ducs(cmd=cmd,params=(b2a(stop),),debug=self._debug)
                cmd = """ DELETE FROM %(trans)s
                          WHERE z_serial < ? """ % tables
                ducs(cmd=cmd,params=(b2a(stop),),debug=self._debug)
                cmd = """ DELETE FROM %(pack)s """ % tables
                ducs(cmd=cmd,debug=self._debug)
            except:
                self._abortSQL()
                import traceback
                traceback.print_exc()
                raise
            else:
                self._commitSQL()
        finally:
            self._lock_release()
                   
    def getSize(self):
        if self._debug:
            print "getSize"
        dcs = self._doCommittedSQL
        tables = self._tables

        cmd = ("SELECT SUM(%(data)s.z_datalen) FROM %(data)s" % tables)
        datasize = int(dcs(cmd=cmd, fetch="one")[0])
        cmd = ("SELECT COUNT(*) FROM %(data)s" % tables)
        numrecs = int(dcs(cmd=cmd, fetch="one")[0])
        cmd = ("SELECT COUNT(*) FROM %(trans)s" % tables)
        numtrans = int(dcs(cmd=cmd, fetch="one")[0])
        data_overhead    = (12 + 12 + 12 + 1 + 32 + 12) * numrecs
        trans_overhead   = (12 + 1 + 255 + 255 + 255) * numtrans
        overhead = data_overhead + trans_overhead
        wag_size = overhead + datasize
        return wag_size # seems to be about 1/2 the diskfile size

    def _testConn(self):
        """ Tests the DB connection and makes sure the tables are there """
        if self._debug:
            print "_testConn"

        tables = self._tables
        dcs = self._doCommittedSQL
        cmds = ("SELECT COUNT(*) FROM %(data)s" % tables,
                "SELECT COUNT(*) FROM %(trans)s" % tables,
                "SELECT COUNT(*) FROM %(pack)s" %tables)

        for cmd in cmds:
            try:
                dcs(cmd = cmd, fetch = "all")
            except kinterbasdb.kinterbasError, value:
                self._closeConn()
                raise RelationalStorageInitError, value
        
    def _closeConn(self):
        if self._debug:
            print "_closeConn"
        self._c.close()

    def _doSQL(self, cmd="", params=[], fetch="", debug=0, dirtyread=0):
        join = string.join
        split = string.split
        if debug:
            print "_doSQL"
        if cmd == "": return
        c = self._c.cursor()
        try:
            cmd = join(split(cmd))
            if debug:
                print cmd, params
            if params:
                c.execute(cmd, params)
            else:
                c.execute(cmd)
            if fetch == "all":
                a = c.fetchall()
                if debug:
                    print "doSQL returned %s rows" % len(a)
                return a
            if fetch == "one":
                a = c.fetchone()
                return a
                
        finally:
            c.close()

    def _doUncommittedSQL(self, *args, **kw):
        debug = self._debug
        _doSQL = self._doSQL
        if debug:
            print "_doUncommittedSQL"
        return apply(_doSQL, args, kw or {})
            
    def _doCommittedSQL(self, *args, **kw):
        debug = self._debug
        _beginSQL = self._beginSQL
        _commitSQL = self._commitSQL
        _doSQL = self._doSQL
        if debug:
            print "_doCommittedSQL"
        dirty = kw.get('dirtyread',None)
        if dirty:
            if debug:
                print 'dirty read tpb in _doCommittedSQL'
            _beginSQL(dirtyread_tpb)
        else:
            _beginSQL()
        try:
            return apply(_doSQL, args, kw or {})
        finally:
            _commitSQL()
            
    def _beginSQL(self, tpb=None):
        debug = self._debug
        _c = self._c
        if debug:
            print "_beginSQL"
        if tpb is None:
            _c.begin() # tpb is Interbase transactional semantics
        else:
            _c.begin(tpb)
            
    def _commitSQL(self):
        debug = self._debug
        _c = self._c
        if debug:
            print "_commitSQL"
        _c.commit()

    def _abortSQL(self):
        debug = self._debug
        _c = self._c
        if debug:
            print "_abortSQL"
        _c.rollback()
        
    def _clear_temp(self):
        _c = self._c
        debug = self._debug
        if debug:
            print "_clear_temp"
        try:
            _c.rollback()
        except kinterbasdb.error:
            pass

    def _abort(self):
        debug = self._debug
        if debug:
            print "_abort"
        self.abortSQL()

def _sql_quote(v):
    return '\"%s\"' % v

def b2a(v):
    b2a_base64 = binascii.b2a_base64
    try:
        return b2a_base64(v)[:-1]
    except:
        if v is None:  return ''
    return v

def a2b(v):
    a2b_base64 = binascii.a2b_base64
    try:
        return a2b_base64(v)
    except:
        if v is None: return ''
    return v


